/*
 *
 *  VMWareSVGA.m
 *
 */

/*	TODO:
	Find a way to avoid artifacts. It seems to occur less often on a lighly loaded machine if the depth is small.
	Look at gamma table stuff and brightness (I suspect that we should not implement gamma, as we are in a virtual device and the host already implement gamma. Note that 8 bit color mode uses the gamma interface and have probably gamma applied twice...)
*/

/*
	Supported modes:
		8 bits grayscale
		8 bits color
		16 bits color
		24 bits color
	How to specify modes:
		Due to apparent limitations in vmware SVGA implementation, the depth is really dependent on the depth of the host environment at driver instanciation time.
		In the case of 8bits, the driver goes into 8 bits color mode, unless the mode choosed is 'BW:8'
	How to specify resolution:
		The driver uses the Height: Width: of the configuration table. If the height/width is specified as zero, the driver uses the height/width of the host at driver instanciation time (so full screen mode resolution matches the host).
	Frequency is ignored
*/

/*
	How to build the driver:
	Drivers can't theorically be developped under OPENSTEP42, mainly because Makefile format have changed. This driver have been made on OPENSTEP42, by hand-editing Makefiles from (NS3.3) Driver Kit example.
	If you open the driver with Project Builder, you should postpone the upgrade, or your PB.project and Makefile will be unusable. Beware that Project Builder writes the PB.project without asking for information when you open the project information inspector (one way to work around this is to change directory ownership to root and revoke write access.)
	Building the driver is easy from the command line (just type make), but is harder with PB, because it want to use gnumake, and will not allow you to change it without writing the broken PB.project file. Working around that involve linking gnumake to make and overriding files while PB is running.

	How to test/develop the driver:
	Developing drivers is challenging as you have to use 2 machines, and crash often. Proper use of VMware will make the process much easier.
	You should use two virtual machines, one for development and one of test. Use the private networking option of vmware to put them on a shared network (say host is 192.168.199.1, dev is 192.168.199.2 and test is 192.168.199.3). Don't bother configuring a netinfo server. Dev machine should export the driver via NFS. In VMWare, remove floppy and CD device from the test machine (no need to probe them at each boot), and set the IDE disk to undoable. Make test machine boot non-graphical and single-user only. This will give you a fast-booting test machine that will never check its disk after a crash.
	When booted single user, load the ethernet driver, configure the ethernet device, mount the file system, copy the driver to its correct location and load the driver. This can be done by a script like:

test # cd /
test # cat g
#!/bin/sh

( echo "y" ; echo "y" ) | driverLoader d=AMDPCnet32NetworkDriver
ifconfig en0 192.168.199.3
mount 192.168.199.2:/General /General
/General/i.sh
test #

The second half of the script is on the dev machine (so you can modify it easily)

dev # cat /General/i.sh
#!/bin/sh

rm -rf /usr/Devices/VMWareSVGA.config
cp -r /General/VMWareSVGA/VMWareSVGA.config /usr/Devices
( echo "y" ; echo "y" ) | driverLoader d=VMWareSVGA
dev #

	So, after boot, you just have to type "/g" to get the driver loaded, and press the power button + discard changes to test the next change.

	Driver structure:
	The whole thing can be done with a quick and dirty 40 lines driver, but I tried to follow the DriverKit recommendation as much as possible.
	The -probe checks that a board is present, sets it to version 2, get the frame buffer address from the board registers, changes the memory range of the driverkit (so, if vmware moves the video ram around, it should still work). There is a problem there, as the PCI configuration return 0x1081 as base register, while, under XFree86, it returns 0x10c0. Neither of those works, the correct base register is 0x1080. I don't understand why this occur and will pose problems when the I/O register will move.
	The initWithDeviceDescription: try to guess a correct mode and maps memory.
 */

#import "VMWareSVGA.h"

#import <driverkit/i386/PCI.h>
#import <driverkit/i386/IOPCIDirectDevice.h>
#import <driverkit/i386/ioPorts.h>

@implementation VMWareSVGA

//	I tried very hard not hard-coding those, but miserabily failed
//	This can be retreived via PCI configuration but the values are wrong
//	Undex XFree86/linux, the retreived values are 0x10c0. I found the one I use by guessing.
//	This will break in the future.
#define INDEX_REG	(0x1080+SVGA_INDEX_PORT)
#define VALUE_REG	(0x1080+SVGA_VALUE_PORT)

//	Index register is the one where the register 'name' must be written
#define SVGA_INDEX_PORT	0x00
//	Value must be read/written here
#define SVGA_VALUE_PORT	0x01

//	Various incarnations of the VMware virtual SVGA board
//	The board report a SVGA_ID_0 version, but let you set the version to 
//	the highest one it support (as january 2002, this is version 2)
#define SVGA_MAGIC	0x900000
#define SVGA_MAKE_ID(ver)	(SVGA_MAGIC<<8 | (ver))
#define SVGA_ID_0	SVGA_MAKE_ID(0)
#define SVGA_ID_1	SVGA_MAKE_ID(1)
#define SVGA_ID_2	SVGA_MAKE_ID(2)

//	The various register (ie: numbers to write into SVGA_INDEX_PORT)
#define SVGA_REG_ID		0x00	//	ID of the board
#define SVGA_REG_ENABLE		0x01	//	When set to 1, goes into SVGA mode
#define SVGA_REG_WIDTH		0x02	//	Width of the screen
#define SVGA_REG_HEIGHT		0x03	//	Height
#define SVGA_REG_MAX_WIDTH	0x04	//	Max width
#define SVGA_REG_MAX_HEIGHT	0x05	//	Max height
#define SVGA_REG_DEPTH		0x06	//	Depth
#define SVGA_REG_BITS_PER_PIXEL	0x07	//	Bpp
#define SVGA_REG_PSEUDOCOLOR	0x08	//	True/false
#define SVGA_REG_RED_MASK	0x09	//	
#define SVGA_REG_GREEN_MASK	0x0a	//	
#define SVGA_REG_BLUE_MASK	0x0b	//	
#define SVGA_REG_BYTES_PER_LINE	0x0c	//	Row bytes
#define SVGA_REG_FB_START	0x0d	//	Start addrs of frame buffer
#define SVGA_REG_FB_OFFSET	0x0e	//	Offset of frame buffer
#define SVGA_REG_FB_MAX_SIZE	0x0f	//	Max size of frame buffer
#define SVGA_REG_FB_SIZE	0x10	//	Size of frame buffer

#define SVGA_PALETTE_BASE	1024

//	Helper routine to read a specific register
//	Example: [self readReg:SVGA_REG_WIDTH]
- (unsigned)readReg:(unsigned)aReg
{	outl( INDEX_REG, aReg );
	return inl( VALUE_REG );
}

//	Helper routine to write to a specific register
//	Example: [self writeReg:SVGA_REG_WIDTH value:1024]
- (void)writeReg:(unsigned)aReg value:(unsigned)aValue
{
	outl( INDEX_REG, aReg );
	outl( VALUE_REG, aValue );
}

/*	I choosed not to trust the memory range from the configuration file because it can be retreived from the virtual device.
	It can also be retreived from the PCI config data (with getPCIConfigData:atRegister:withDeviceDescription:) but the result is wrong (?)
	So, I get it from the device and manually set the range.
	I had to implement +(BOOL)probe, as hacking ranges have to be done before initWithDeviceDescription:
*/
+ (BOOL)probe:(IODeviceDescription *)aDeviceDescription
{	IORange theFBRange;
	VMWareSVGA *theDevice;
	static unsigned ids[] = { SVGA_ID_2, SVGA_ID_1, SVGA_ID_0 };
	int theIdCount = sizeof(ids)/sizeof(ids[0]);
	unsigned theVersion;
	int i;

	IOLog( "VMWareSVGA: Probing\n" );

	//	First, we check that we really have a VMWare SVGA board
	outl( INDEX_REG, SVGA_REG_ID );
	theVersion = inl( VALUE_REG );
	if (theVersion!=SVGA_ID_0 && theVersion!=SVGA_ID_2)	//	When rebooting, vmware don't reset version (!)
	{	IOLog( "VMWareSVGA: did not found device at %0x (INDEX_REG is wrong)\n", INDEX_REG );
		IOLog( "VMWareSVGA: got %0x\n", theVersion );
		return NO;
	}

	//	Second, we try to set it into version 2
	for (i=0;i!=theIdCount;i++)
	{	outl( INDEX_REG, SVGA_REG_ID );
		outl( VALUE_REG, ids[i] );
		outl( INDEX_REG, SVGA_REG_ID );
		theVersion = inl( VALUE_REG );
		if (ids[i]!=theVersion)
			IOLog( "  Device does not support %0x [%0x] -- (???)\n", ids[i], theVersion );
		else
			break;
	}
	if (i==theIdCount)
	{	IOLog( "  %s: Cannot initialize device to SVGA_ID_[0-2]\n", [self name] );	
		return NO;
	}
	IOLog( "  VMWare SVGA version %0x probed\n", theVersion );


	//	Third, we correct what the DriverKit thinks about range used by the driver (we should add INDEX_REG too, but this doesn't prevent the driver to work)
	outl( INDEX_REG, SVGA_REG_FB_START );
	theFBRange.start = inl( VALUE_REG );
	outl( INDEX_REG, SVGA_REG_FB_SIZE );
	theFBRange.size = inl( VALUE_REG );
	[aDeviceDescription setMemoryRangeList:&theFBRange num:1];

	//	Fourth, we allocate the device
	theDevice = [[self alloc] initFromDeviceDescription:aDeviceDescription];
	if (!theDevice)
		return NO;

	//	Fifth we register it
	[theDevice registerDevice];

	return YES;
}

//	Helpers due to lack of stdlib in the kernel
extern char *strstr(const char *s1, const char *s2);
static int my_atoi( const char *s )
{	int n = 0;
	while (*s==' ')
		s++;
	while (*s>='0' && *s<='9')
		n = n*10+*s++-'0';
	return n;
}
static int parse_int( const char *s1, const char *s2 )
{	s1 = strstr( s1, s2 );
	if (!s1)
		return 0;
	return my_atoi( s1+strlen(s2) );
}

/*	There are two problems with the config table approach of the driver kit:
	1/ The vmware SVGA device theorically support many different weird resolutions
		In a 'classic' driver, we would need to put a list of those is Display.modes _and_ make a static list of usable modes in the code. This would, IMHO, suck
	2/ Contrary to a real board, the vmware virtual device cannot change bit depth (I consider this as a bug), so we can not use the real depth choosed by the user (if we did it would mean that one would have to reboot OPENSTEP, use configure to change the depth and reboot when changing the real depth of the underying OS)
	The result is that we only have a single mode that we dynamically construct from the device description, overriding the fields we cannot honor with more plausible values. In addition we support zero width/height as a way to use the width/height of the underlying OS (for nicer full-screen mode)
 */
- initFromDeviceDescription:deviceDescription
{	IODisplayInfo theDisplayInfo;
	const char *theModeString;
	BOOL thefBW = NO;

	if (![super initFromDeviceDescription:deviceDescription])
		return nil;

		//	Getting user specs
	theModeString = [[[self deviceDescription] configTable] valueForStringKey:"Display Mode"];
	if (theModeString)
	{	int theWidth;
		int theHeight;

		theWidth = parse_int( theModeString, "Width:" );
		if (theWidth)
		{	IOLog( "  forcing width from %d to %d ", [self readReg:SVGA_REG_WIDTH], theWidth );
			[self writeReg:SVGA_REG_WIDTH value:theWidth];
			IOLog( "got %d\n", [self readReg:SVGA_REG_WIDTH] );
		}

		theHeight = parse_int( theModeString, "Height:" );
		if (theHeight)
		{	IOLog( "  forcing height from %d to %d ", [self readReg:SVGA_REG_HEIGHT], theHeight );
			[self writeReg:SVGA_REG_HEIGHT value:theHeight];
			IOLog( "got %d\n", [self readReg:SVGA_REG_HEIGHT] );
		}

			//	Check for BW mode
		if (strstr(theModeString,"BW"))
			thefBW = YES;
	}

		//	We correct the mode to match vmware's one
		//	Building the display info
	theDisplayInfo.width = [self readReg:SVGA_REG_WIDTH];
	theDisplayInfo.height = [self readReg:SVGA_REG_HEIGHT];
	theDisplayInfo.totalWidth = [self readReg:SVGA_REG_WIDTH];
	theDisplayInfo.rowBytes = [self readReg:SVGA_REG_BYTES_PER_LINE];
		//	Hard coded. Don't know how to pass it to vmware anyway.
	theDisplayInfo.refreshRate = 70;
	theDisplayInfo.flags = 0;
	switch ([self readReg:SVGA_REG_DEPTH])
	{
		case 8:
			theDisplayInfo.bitsPerPixel = IO_8BitsPerPixel;
			if (thefBW)
			{	int i;

				theDisplayInfo.colorSpace = IO_OneIsWhiteColorSpace;
				strcpy( theDisplayInfo.pixelEncoding, "WWWWWWWW" );
				for (i = 0; i < 256; i++)
				{	[self writeReg:SVGA_PALETTE_BASE + i * 3 + 0 value:i];
					[self writeReg:SVGA_PALETTE_BASE + i * 3 + 1 value:i];
					[self writeReg:SVGA_PALETTE_BASE + i * 3 + 2 value:i];
				}
			}
			else
			{	theDisplayInfo.colorSpace = IO_8BitsPerPixel;
				strcpy( theDisplayInfo.pixelEncoding, "PPPPPPPP" );
					//	The window server will send us the palette to use
				theDisplayInfo.flags |= IO_DISPLAY_HAS_TRANSFER_TABLE;
			}
			break;
		case 24:
			theDisplayInfo.bitsPerPixel = IO_24BitsPerPixel;
			theDisplayInfo.colorSpace = IO_RGBColorSpace;
			strcpy( theDisplayInfo.pixelEncoding, "--------RRRRRRRRGGGGGGGGBBBBBBBB" );
			break;
		case 15:
			theDisplayInfo.bitsPerPixel = IO_15BitsPerPixel;
			theDisplayInfo.colorSpace = IO_RGBColorSpace;
			strcpy( theDisplayInfo.pixelEncoding, "-RRRRRGGGGGBBBBB" );
			break;
		default:
			//	I suppose that values like 32 and 16 are trivial to deal with
			IOLog( "%s: unsupported BPP of %d, please report to f5426@almonde.com\n", [self name], [self readReg:SVGA_REG_DEPTH] );
			[self free];
			return nil;
	}
	theDisplayInfo.memorySize = [self readReg:SVGA_REG_FB_SIZE];
	theDisplayInfo.screenWidth = [self readReg:SVGA_REG_WIDTH];
	theDisplayInfo.screenHeight = [self readReg:SVGA_REG_HEIGHT];

	IOLog( "  %s: %dx%d/%d [%s]\n", [self name], [self readReg:SVGA_REG_WIDTH], [self readReg:SVGA_REG_HEIGHT], [self readReg:SVGA_REG_DEPTH], [self readReg:SVGA_REG_DEPTH]==8&&thefBW?"grayscale":"color" );

	theDisplayInfo.flags |= IO_DISPLAY_CACHE_OFF;
	[self mapMemoryRange:0 
		to:(vm_address_t *)&(theDisplayInfo.frameBuffer)
		findSpace:YES cache:IO_CacheOff];	//	Tried several values, none avoid the ugly artifacts

	if (theDisplayInfo.frameBuffer==0)
	{	IOLog( "  %s: Cannot map framebuffer\n", [self name] );
		[super free];
		return nil;
	}

		//	Tell the superclass about our choices
	*[self displayInfo] = theDisplayInfo;

	return self;
}

- (void)enterLinearMode
{
	[self writeReg:SVGA_REG_ENABLE value:1];
	[super enterLinearMode];
}

- (void)revertToVGAMode
{
	[self writeReg:SVGA_REG_ENABLE value:0];
	[super revertToVGAMode];
}

//	Very hackish. The Window server uses the transfer table to send the palette to use in PPPPPPPP mode (!)
//	Unfortunately, this is also the interface used for gamma correction, so the 8 bit color mode is gamma corrected
- setTransferTable:(const unsigned int *)table count:(int)numEntries
{	int i;

	//	We know that we are in 8 bits RGB mode
	for (i = 0; i < 256; i++)
	{	[self writeReg:SVGA_PALETTE_BASE + i * 3 + 0 value:(table[i]>>24)&0xFF];
		[self writeReg:SVGA_PALETTE_BASE + i * 3 + 1 value:(table[i]>>16)&0xFF];
		[self writeReg:SVGA_PALETTE_BASE + i * 3 + 2 value:(table[i]>>8)&0xFF];
	}

	return self;
}

@end
